@@ -22,6 +22,7 @@
 #include "wsrep_params.hpp"
 #include "gu_event_service.hpp"
 #include "wsrep_config_service.h"
+#include "wsrep_node_isolation.h"
 
 #include <cassert>
 
@@ -534,12 +535,8 @@ galera_terminate_trx(wsrep_t*           const gh,
         trx->set_state(TrxHandle::S_MUST_ABORT);
         trx->set_state(TrxHandle::S_ABORTING);
     }
-    wsrep_status_t retval(repl->send(*trx, meta));
-    if (retval == WSREP_OK)
-    {
-        retval = galera_sync_wait(gh, NULL, -1, NULL);
-    }
-    return retval;
+
+    return repl->terminate_trx(*trx, meta);
 }
 
 extern "C"
@@ -1002,14 +999,25 @@ wsrep_status_t galera_append_key(wsrep_t*           const gh,
 
     try
     {
+        int const proto_ver(repl->trx_proto_ver());
         TrxHandleLock lock(*trx);
-        for (size_t i(0); i < keys_num; ++i)
+
+        if (keys_num > 0)
         {
-            galera::KeyData k (repl->trx_proto_ver(),
-                               keys[i].key_parts,
-                               keys[i].key_parts_num,
-                               key_type,
-                               copy);
+            for (size_t i(0); i < keys_num; ++i)
+            {
+                galera::KeyData const k(proto_ver,
+                                        keys[i].key_parts,
+                                        keys[i].key_parts_num,
+                                        key_type,
+                                        copy);
+                gu_trace(trx->append_key(k));
+            }
+        }
+        else if (proto_ver >= 6)
+        {
+            /* Append server-level key (matches every trx)*/
+            galera::KeyData const k(proto_ver, key_type);
             gu_trace(trx->append_key(k));
         }
         retval = WSREP_OK;
@@ -1915,3 +1923,22 @@ wsrep_status_t wsrep_init_membership_service_v1(
 extern "C" void wsrep_deinit_membership_service_v1()
 {
 }
+
+/*
+ * This function may be called from signal handler, so make sure that
+ * only 'safe' system calls and library functions are used. See
+ * https://pubs.opengroup.org/onlinepubs/009695399/functions/xsh_chap02_04.html
+ */
+extern "C" enum wsrep_node_isolation_result
+wsrep_node_isolation_mode_set_v1(enum wsrep_node_isolation_mode mode)
+{
+    if (mode < WSREP_NODE_ISOLATION_NOT_ISOLATED
+        || mode > WSREP_NODE_ISOLATION_FORCE_DISCONNECT)
+    {
+        return WSREP_NODE_ISOLATION_INVALID_VALUE;
+    }
+    gu::gu_asio_node_isolation_mode = mode;
+    return WSREP_NODE_ISOLATION_SUCCESS;
+}
+
+
