@@ -256,6 +256,11 @@ std::ostream& gu::operator<<(std::ostream& os, const gu::AsioErrorCode& ec)
     return (os << ec.message());
 }
 
+gu::AsioErrorCode gu::AsioErrorCode::make_eof()
+{
+  return {asio::error::misc_errors::eof, gu_asio_misc_category};
+}
+
 bool gu::AsioErrorCode::is_eof() const
 {
     return (category_ &&
@@ -314,7 +319,16 @@ namespace
 
         std::string get_password() const
         {
-            std::string   file(conf_.get(gu::conf::ssl_password_file));
+            std::string   file;
+            try {
+                file = conf_.get(gu::conf::ssl_password_file);
+            }
+            catch (const gu::NotSet&)
+            {
+                gu_throw_error(EINVAL)
+                << gu::conf::ssl_password_file << " is required";
+            }
+
             std::ifstream ifs(file.c_str(), std::ios_base::in);
 
             if (ifs.good() == false)
@@ -327,6 +341,7 @@ namespace
             std::getline(ifs, ret);
             return ret;
         }
+
     private:
         const gu::Config& conf_;
     };
@@ -391,7 +406,7 @@ std::string gu::extra_error_info(const gu::AsioErrorCode& ec)
 
 static SSL_CTX* native_ssl_ctx(asio::ssl::context& context)
 {
-#if ASIO_VERSION < 101601
+#if ASIO_VERSION < 101401
     return context.impl();
 #else
     return context.native_handle();
@@ -536,7 +551,8 @@ void gu::ssl_register_params(gu::Config& conf)
              gu::Config::Flag::type_bool);
     conf.add(gu::conf::ssl_compression,
              gu::Config::Flag::read_only |
-             gu::Config::Flag::type_bool);
+             gu::Config::Flag::type_bool |
+             gu::Config::Flag::deprecated);
     conf.add(gu::conf::ssl_key,
              gu::Config::Flag::read_only);
     conf.add(gu::conf::ssl_cert,
@@ -561,7 +577,7 @@ void gu::ssl_param_set(const std::string& key, const std::string& val,
         {
             try
             {
-#if ASIO_VERSION < 101601
+#if ASIO_VERSION < 101401
                 asio::io_service io_service;
                 asio::ssl::context ctx(io_service, asio::ssl::context::sslv23);
 #else
@@ -600,20 +616,26 @@ void gu::ssl_init_options(gu::Config& conf)
         conf.set(conf::ssl_cipher, cipher_list);
 
         // compression
-        bool compression(conf.get(conf::ssl_compression, true));
-        if (compression == false)
+        try
         {
-            log_info << "disabling SSL compression";
-            sk_SSL_COMP_zero(SSL_COMP_get_compression_methods());
+            (void) conf.get(conf::ssl_compression);
+            // warn the user if socket.ssl_compression is set explicitly
+            log_warn << "SSL compression is not effective. The option "
+                     << conf::ssl_compression << " is deprecated and "
+                     << "will be removed in future releases.";
         }
-        conf.set(conf::ssl_compression, compression);
-
+        catch (NotSet&)
+        {
+            // this is a desirable situation
+        }
+        log_info << "not using SSL compression";
+        sk_SSL_COMP_zero(SSL_COMP_get_compression_methods());
 
         // verify that asio::ssl::context can be initialized with provided
         // values
         try
         {
-#if ASIO_VERSION < 101601
+#if ASIO_VERSION < 101401
             asio::io_service io_service;
             asio::ssl::context ctx(io_service, asio::ssl::context::sslv23);
 #else
@@ -929,3 +951,7 @@ void gu::deinit_allowlist_service_v1()
     --gu_allowlist_service_usage;
     if (gu_allowlist_service_usage == 0) gu_allowlist_service = 0;
 }
+
+std::atomic<enum wsrep_node_isolation_mode> gu::gu_asio_node_isolation_mode{
+    WSREP_NODE_ISOLATION_NOT_ISOLATED
+};
