@@ -37,9 +37,10 @@ static bool check_tcp_uri(const gu::URI& uri)
             uri.get_scheme() == gu::scheme::ssl);
 }
 
-static std::string get_scheme(bool use_ssl)
+static std::string get_scheme(const gcomm::Protonet& pnet,
+                              bool use_ssl, bool dynamic_socket)
 {
-    if (use_ssl == true)
+    if ((use_ssl == true || pnet.tls_service_enabled()) && not dynamic_socket)
     {
         return gu::scheme::ssl;
     }
@@ -81,13 +82,22 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri,
                           param<int>(conf_, uri, Conf::GMCastSegment, "0"),
                           0, 255)),
     my_uuid_      (my_uuid ? *my_uuid : UUID(0, 0)),
-    use_ssl_      (param<bool>(conf_, uri, gu::conf::use_ssl, "false")),
+    dynamic_socket_ (conf_.has(gu::conf::socket_dynamic) ?
+                     param<bool>(conf_, uri, gu::conf::socket_dynamic,
+                                 "false") :
+                     false),
+#ifdef GALERA_HAVE_SSL
+    use_ssl_      (param<bool>(conf_, uri, gu::conf::use_ssl, "false") ||
+                   pnet_.tls_service_enabled()),
+#else
+    use_ssl_(),
+#endif // GALERA_HAVE_SSL
     // @todo: technically group name should be in path component
     group_name_   (param<std::string>(conf_, uri, Conf::GMCastGroup, "")),
     listen_addr_  (
         param<std::string>(
             conf_, uri, Conf::GMCastListenAddr,
-            get_scheme(use_ssl_) + "://0.0.0.0")), // how to make it IPv6 safe?
+            get_scheme(pnet_, use_ssl_, dynamic_socket_) + "://0.0.0.0")), // how to make it IPv6 safe?
     initial_addrs_(),
     mcast_addr_   (param<std::string>(conf_, uri, Conf::GMCastMCastAddr, "")),
     bind_ip_      (""),
@@ -95,7 +105,7 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri,
                        Conf::GMCastMCastTTL,
                        param<int>(conf_, uri, Conf::GMCastMCastTTL, "1"),
                        1, 256)),
-    listener_     (0),
+    listener_     (),
     mcast_        (),
     pending_addrs_(),
     remote_addrs_ (),
@@ -143,7 +153,7 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri,
     catch (gu::Exception&)
     {
         /* most probably no scheme, try to append one and see if it succeeds */
-        listen_addr_ = uri_string(get_scheme(use_ssl_), listen_addr_);
+        listen_addr_ = uri_string(get_scheme(pnet_, use_ssl_, dynamic_socket_), listen_addr_);
         gu_trace(gu::URI uri(listen_addr_));
     }
 
@@ -188,11 +198,10 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri,
 
     listen_addr_ = gu::net::resolve(listen_addr_).to_string();
     // resolving sets scheme to tcp, have to rewrite for ssl
-    if (use_ssl_ == true)
+    if ((use_ssl_ == true || pnet_.tls_service_enabled()) && not dynamic_socket_)
     {
         listen_addr_.replace(0, 3, gu::scheme::ssl);
     }
-
     std::set<std::string>::iterator iaself(initial_addrs_.find(listen_addr_));
     if (iaself != initial_addrs_.end())
     {
@@ -205,7 +214,7 @@ gcomm::GMCast::GMCast(Protonet& net, const gu::URI& uri,
     {
         try
         {
-            port = uri_.get_option(Conf::GMCastMCastPort);
+            port = param<std::string>(conf_, uri_, Conf::GMCastMCastPort, port);
         }
         catch (gu::NotFound&) {}
 
@@ -274,7 +283,7 @@ void gcomm::GMCast::set_initial_addr(const gu::URI& uri)
             }
         }
 
-        std::string initial_uri = uri_string(get_scheme(use_ssl_), host, port);
+        std::string initial_uri = uri_string(get_scheme(pnet_, use_ssl_, dynamic_socket_), host, port);
         std::string initial_addr;
         try
         {
@@ -285,9 +294,9 @@ void gcomm::GMCast::set_initial_addr(const gu::URI& uri)
             log_warn << "Failed to resolve " << initial_uri;
             continue;
         }
-
         // resolving sets scheme to tcp, have to rewrite for ssl
-        if (use_ssl_ == true)
+        if ((use_ssl_ == true || pnet_.tls_service_enabled()) &&
+            not dynamic_socket_)
         {
             initial_addr.replace(0, 3, gu::scheme::ssl);
         }
@@ -373,8 +382,7 @@ void gcomm::GMCast::close(bool force)
 
     gcomm_assert(listener_ != 0);
     listener_->close();
-    delete listener_;
-    listener_ = 0;
+    listener_.reset();
 
     segment_map_.clear();
     for (ProtoMap::iterator
@@ -626,8 +634,7 @@ void gcomm::GMCast::gmcast_forget(const UUID& uuid,
                     erase_proto(pi);
                 }
             }
-            ae.set_max_retries(0);
-            ae.set_retry_cnt(1);
+            disable_reconnect(*ai);
             gu::datetime::Date now(gu::datetime::Date::monotonic());
             // Don't reduce next reconnect time if it is set greater than
             // requested
@@ -719,12 +726,7 @@ void gcomm::GMCast::handle_established(Proto* est)
         return;
     }
 
-    // send_up(Datagram(), p->remote_uuid());
-
-    // init retry cnt to -1 to avoid unnecessary logging at first attempt
-    // max retries will be readjusted in handle stable view
-    AddrList::value(i).set_retry_cnt(-1);
-    AddrList::value(i).set_max_retries(max_initial_reconnect_attempts_);
+    enable_reconnect(*i);
 
     // Cleanup all previously established entries with same
     // remote uuid. It is assumed that the most recent connection
@@ -965,11 +967,7 @@ void gcomm::GMCast::update_addresses()
 
                     AddrEntry& ae(AddrList::value(pi));
 
-                    // init retry cnt to -1 to avoid unnecessary logging
-                    // at first attempt
-                    // max retries will be readjusted in handle stable view
-                    ae.set_retry_cnt(-1);
-                    ae.set_max_retries(max_initial_reconnect_attempts_);
+                    enable_reconnect(*pi);
 
                     // Add some randomness for first reconnect to avoid
                     // simultaneous connects
@@ -1106,6 +1104,30 @@ void gcomm::GMCast::reconnect()
     }
 }
 
+void gcomm::GMCast::disable_reconnect(AddrList::value_type& entry)
+{
+    log_debug << "Disabling reconnect for " << entry.first;
+    entry.second.set_max_retries(0);
+    entry.second.set_retry_cnt(1);
+}
+
+void gcomm::GMCast::enable_reconnect(AddrList::value_type& entry)
+{
+    if (entry.second.retry_cnt() == -1)
+    {
+        return;
+    }
+    log_debug << "Enabling reconnect for " << entry.first;
+    /* Initialize retry cnt to -1 in order to avoid unnecessary logging
+     * at the first connect attempt.
+     * Initial reconnect attempts is limited to terminate reconnection
+     * attempts early if the peer becomes unresponsive. The max retries
+     * is adjusted into higher value in handle_stable_view() once it is
+     * known that the peer has become part of the stable view. */
+    entry.second.set_retry_cnt(-1);
+    entry.second.set_max_retries(max_initial_reconnect_attempts_);
+}
+
 namespace
 {
     class CmpUuidCounts
@@ -1227,15 +1249,15 @@ void gcomm::GMCast::check_liveness()
         if (ae.retry_cnt()             <= ae.max_retries() &&
             live_uuids.find(ae.uuid()) == live_uuids.end())
         {
-            // log_info << self_string()
-            // << " missing live proto entry for " << ae.uuid();
+            log_debug << self_string()
+                      << " missing live proto entry for " << ae.uuid();
             nonlive_uuids.insert(ae.uuid());
             nonlive_peers += AddrList::key(i) + " ";
             should_relay = true;
         }
         else if (ae.last_connect() + peer_timeout_ > now)
         {
-            log_debug << "continuing relaying for "
+            log_debug << "continuing relaying to " << i->first << " for "
                       << (ae.last_connect() + peer_timeout_ - now);
             should_relay = true;
         }
@@ -1245,9 +1267,12 @@ void gcomm::GMCast::check_liveness()
     {
         if (relaying_ == false)
         {
-            log_info << self_string()
-                     << " turning message relay requesting on, nonlive peers: "
-                     << nonlive_peers;
+            if (not nonlive_uuids.empty())
+            {
+                log_info << self_string()
+                         << " turning message relay requesting on, nonlive peers: "
+                         << nonlive_peers;
+            }
             relaying_ = true;
         }
         relay_set_.clear();
@@ -1316,7 +1341,7 @@ void gcomm::GMCast::send(const RelayEntry& re, int segment, gcomm::Datagram& dg)
         log_debug << "failed to send to " << re.socket->remote_addr()
                   << ": (" << err << ") " << strerror(err);
     }
-    else
+    else if (re.proto)
     {
         re.proto->set_send_tstamp(gu::datetime::Date::monotonic());
     }
@@ -1774,6 +1799,15 @@ void gcomm::GMCast::handle_stable_view(const View& view)
     }
 }
 
+void gcomm::GMCast::handle_allow_connect(const UUID& uuid)
+{
+    auto it = std::find_if(remote_addrs_.begin(), remote_addrs_.end(),
+                            AddrListUUIDCmp(uuid));
+    if (it != remote_addrs_.end())
+    {
+        enable_reconnect(*it);
+    }
+}
 
 void gcomm::GMCast::handle_evict(const UUID& uuid)
 {
@@ -1799,15 +1833,13 @@ void gcomm::GMCast::add_or_del_addr(const std::string& val)
     if (val.compare(0, 4, "add:") == 0)
     {
         gu::URI uri(val.substr(4));
-        std::string addr(gu::net::resolve(uri_string(get_scheme(use_ssl_),
+        std::string addr(gu::net::resolve(uri_string(get_scheme(pnet_, use_ssl_, dynamic_socket_),
                                                      uri.get_host(),
                                                      uri.get_port())).to_string());
         log_info << "inserting address '" << addr << "'";
         insert_address(addr, UUID(), remote_addrs_);
         AddrList::iterator ai(remote_addrs_.find(addr));
-        AddrList::value(ai).set_max_retries(
-            max_initial_reconnect_attempts_);
-        AddrList::value(ai).set_retry_cnt(-1);
+        enable_reconnect(*ai);
     }
     else if (val.compare(0, 4, "del:") == 0)
     {
@@ -1827,8 +1859,7 @@ void gcomm::GMCast::add_or_del_addr(const std::string& val)
                 }
             }
             AddrEntry& ae(AddrList::value(ai));
-            ae.set_max_retries(0);
-            ae.set_retry_cnt(1);
+            disable_reconnect(*ai);
             ae.set_next_reconnect(gu::datetime::Date::monotonic() + time_wait_);
             update_addresses();
         }
