@@ -1399,7 +1399,7 @@ _close(gcs_conn_t* conn, bool join_recv_thread)
             assert (GCS_CONN_CLOSED == conn->state);
         }
 
-        gu_info ("Closing replication queue.");
+        gu_info ("Closing send queue.");
         struct gcs_repl_act** act_ptr;
         /* At this point (state == CLOSED) no new threads should be able to
          * queue for repl (check gcs_repl()), and recv thread is joined, so no
@@ -1422,7 +1422,7 @@ _close(gcs_conn_t* conn, bool join_recv_thread)
         /* wake all gcs_recv() threads () */
         // FIXME: this can block waiting for applicaiton threads to fetch all
         // items. In certain situations this can block forever. Ticket #113
-        gu_info ("Closing slave action queue.");
+        gu_info ("Closing receive queue.");
         gu_fifo_close (conn->recv_q);
     }
 
@@ -1494,7 +1494,14 @@ static void *gcs_recv_thread (void *arg)
         {
             ret = gcs_handle_actions (conn, rcvd);
 
-            if (gu_unlikely(ret < 0)) {         // error
+            if (gu_unlikely(ret <= 0 && GCS_ACT_COMMIT_CUT == rcvd.act.type))
+            {
+                /* Commit cut will be discarded, the buffer needs to be
+                 * freed */
+                ::free(const_cast<void*>(rcvd.act.buf));
+            }
+            if (gu_unlikely(ret < 0))
+            { // error
                 gu_debug ("gcs_handle_actions returned %d: %s",
                           ret, strerror(-ret));
                 break;
@@ -2232,9 +2239,25 @@ gcs_set_last_applied (gcs_conn_t* conn, const gu::GTID& gtid)
 
     long ret = gcs_sm_enter (conn->sm, &cond, false, false);
 
-    if (!ret) {
-        ret = gcs_core_set_last_applied (conn->core, gtid);
-        gcs_sm_leave (conn->sm);
+    if (ret)
+    {
+        log_info << "Unable to report last applied write-set to "
+                 << "cluster. Will try later. "
+                 << "(gcs_sm_enter(): " << -ret
+                 << " seqno: " << gtid.seqno() << ")";
+    }
+    else
+    {
+        log_debug << "Sending last applied seqno: " << gtid.seqno();
+        ret = gcs_core_set_last_applied(conn->core, gtid);
+        gcs_sm_leave(conn->sm);
+        if (ret < 0)
+        {
+            log_info << "Unable to report last applied write-set to "
+                     << "cluster. Will try later. "
+                     << "(gcs_core_set_last_applied(): " << -ret
+                     << " seqno: " << gtid.seqno() << ")";
+        }
     }
 
     gu_cond_destroy (&cond);
@@ -2255,11 +2278,19 @@ gcs_vote (gcs_conn_t* const conn, const gu::GTID& gtid, uint64_t const code,
     if (gcs_proto_ver(conn) < 1)
     {
         assert(code != 0); // should be here only our own initiative
-        log_error << "Not all group members support inconsistency voting. "
-                  << "Reverting to old behavior: abort on error.";
+        log_info << "Not all group members support inconsistency voting. "
+                 << "Reverting to old behavior: abort on error.";
         return 1; /* no voting with old protocol */
     }
 
+    if (conn->state >= GCS_CONN_JOINER)
+    {
+        assert(code != 0); // should be here only our own initiative
+        log_info << "Can't vote when not at least JOINED. "
+                 << "Assuming inconsistency. Full SST is required";
+        return 1; /* Error applying IST event */
+    }
+
     int const err(gu_mutex_lock(&conn->vote_lock_));
     if (gu_unlikely(0 != err))
     {
