@@ -24,10 +24,10 @@
 #include <gu_backtrace.hpp>
 #include <gu_throw.hpp>
 #include <gu_logger.hpp>
-#include <gu_barrier.hpp>
 #include <gu_thread.hpp>
 
 #include <deque>
+#include <future>
 
 using namespace std;
 using namespace gu;
@@ -91,7 +91,9 @@ private:
 
 public:
 
-    RecvBuf() : mutex_(), cond_(), queue_(), waiting_(false) { }
+    RecvBuf() : mutex_(gu::get_mutex_key(gu::GU_MUTEX_KEY_GCS_GCOMM_RECV_BUF)),
+                cond_(gu::get_cond_key(gu::GU_COND_KEY_GCS_GCOMM_RECV_BUF)),
+                queue_(), waiting_(false) { }
 
     void push_back(const RecvBufData& p)
     {
@@ -148,107 +150,31 @@ public:
         uuid_(),
         thd_(),
         schedparam_(conf_.get(gcomm_thread_schedparam_opt)),
-        barrier_(2),
         uri_(u),
         net_(Protonet::create(conf_)),
         tp_(0),
-        mutex_(),
+        mutex_(gu::get_mutex_key(gu::GU_MUTEX_KEY_GCS_GCOMM_CONN)),
         refcnt_(0),
         terminated_(false),
         error_(0),
         recv_buf_(),
-        current_view_()
+        current_view_(),
+        connect_task_()
     {
         log_info << "backend: " << net_->type();
     }
 
     ~GCommConn()
     {
+        delete tp_;
         delete net_;
     }
 
     const gcomm::UUID& get_uuid() const { return uuid_; }
 
-    static void* run_fn(void* arg)
-    {
-        static_cast<GCommConn*>(arg)->run();
-        return 0;
-    }
-
     void connect(bool) { }
 
-
-    void connect(const string& channel, bool const bootstrap)
-    {
-        if (tp_ != 0)
-        {
-            gu_throw_fatal << "backend connection already open";
-        }
-
-
-        error_ = ENOTCONN;
-        int err;
-        if ((err = gu_thread_create(&thd_, 0, &run_fn, this)) != 0)
-        {
-            gu_throw_error(err) << "Failed to create thread";
-        }
-
-        // Helper to call barrier_.wait() when goes out of scope
-        class StartBarrier
-        {
-        public:
-            StartBarrier(Barrier& barrier) : barrier_(barrier) { }
-            ~StartBarrier()
-            {
-                barrier_.wait();
-            }
-        private:
-            Barrier& barrier_;
-        } start_barrier(barrier_);
-
-        thread_set_schedparam(thd_, schedparam_);
-        log_info << "gcomm thread scheduling priority set to "
-                 << thread_get_schedparam(thd_) << " ";
-
-        uri_.set_option("gmcast.group", channel);
-        tp_ = Transport::create(*net_, uri_);
-        gcomm::connect(tp_, this);
-
-        if (bootstrap)
-        {
-            log_info << "gcomm: bootstrapping new group '" << channel << '\'';
-        }
-        else
-        {
-            string peer;
-            URI::AuthorityList::const_iterator i, i_next;
-            for (i = uri_.get_authority_list().begin();
-                 i != uri_.get_authority_list().end(); ++i)
-            {
-                i_next = i;
-                ++i_next;
-                string host;
-                string port;
-                try { host = i->host(); } catch (NotSet&) { }
-                try { port = i->port(); } catch (NotSet&) { }
-                peer += host != "" ? host + ":" + port : "";
-                if (i_next != uri_.get_authority_list().end())
-                {
-                    peer += ",";
-                }
-            }
-            log_info << "gcomm: connecting to group '" << channel
-                     << "', peer '" << peer << "'";
-        }
-
-        tp_->connect(bootstrap);
-
-        uuid_ = tp_->uuid();
-
-        error_ = 0;
-
-        log_info << "gcomm: connected";
-    }
+    void connect(string channel, bool const bootstrap);
 
     void close(bool force = false)
     {
@@ -359,11 +285,12 @@ private:
 
     void unref() { }
 
+    void print_connect_diag(const std::string&, bool boostrap) const;
+
     gu::Config&       conf_;
     gcomm::UUID       uuid_;
     gu_thread_t       thd_;
     ThreadSchedparam  schedparam_;
-    Barrier           barrier_;
     URI               uri_;
     Protonet*         net_;
     Transport*        tp_;
@@ -373,8 +300,89 @@ private:
     int               error_;
     RecvBuf           recv_buf_;
     View              current_view_;
+    std::packaged_task<void()> connect_task_;
 };
 
+extern "C"
+void* run_fn(void* arg)
+{
+    static_cast<GCommConn*>(arg)->run();
+    gu_thread_exit(0);
+}
+
+void GCommConn::print_connect_diag(const std::string& channel,
+                                   bool const bootstrap) const
+{
+    if (bootstrap)
+    {
+        log_info << "gcomm: bootstrapping new group '" << channel << '\'';
+    }
+    else
+    {
+        string peer;
+        URI::AuthorityList::const_iterator i, i_next;
+        for (i = uri_.get_authority_list().begin();
+             i != uri_.get_authority_list().end(); ++i)
+        {
+            i_next = i;
+            ++i_next;
+            string host;
+            string port;
+            try { host = i->host(); } catch (NotSet&) { }
+            try { port = i->port(); } catch (NotSet&) { }
+            peer += host != "" ? host + ":" + port : "";
+            if (i_next != uri_.get_authority_list().end())
+            {
+                peer += ",";
+            }
+        }
+        log_info << "gcomm: connecting to group '" << channel
+                 << "', peer '" << peer << "'";
+    }
+}
+
+void GCommConn::connect(string channel, bool const bootstrap)
+{
+    if (tp_ != 0)
+    {
+        gu_throw_fatal << "backend connection already open";
+    }
+
+    /* This task is invoked at the very beginning of
+     * run() method. */
+    connect_task_ = std::packaged_task<void()>{
+        [this, channel, bootstrap]()
+        {
+            gcomm::Critical<Protonet> crit(*net_);
+            uri_.set_option("gmcast.group", channel);
+            tp_ = Transport::create(*net_, uri_);
+            gcomm::connect(tp_, this);
+            print_connect_diag(channel, bootstrap);
+            tp_->connect(bootstrap);
+            uuid_ = tp_->uuid();
+            error_ = 0;
+            log_info << "gcomm: connected";
+        }
+    };
+
+    auto future = connect_task_.get_future();
+
+    error_ = ENOTCONN;
+    int err;
+    if ((err = gu_thread_create(gu::get_thread_key(gu::GU_THREAD_KEY_GCS_GCOMM),
+                                &thd_, run_fn, this))
+        != 0)
+    {
+        gu_throw_error(err) << "Failed to create thread";
+    }
+
+    thread_set_schedparam(thd_, schedparam_);
+    log_info << "gcomm thread scheduling priority set to "
+             << thread_get_schedparam(thd_) << " ";
+
+    /* Will throw if an exception was thrown in connect_task. */
+    future.get();
+}
 
 void
 GCommConn::handle_up(const void* id, const Datagram& dg, const ProtoUpMeta& um)
@@ -414,8 +422,8 @@ GCommConn::handle_up(const void* id, const Datagram& dg, const ProtoUpMeta& um)
 
 void GCommConn::run()
 {
-    barrier_.wait();
-    if (error_ != 0) gu_thread_exit(0);
+    connect_task_();
+    if (error_ != 0) return;
 
     while (true)
     {
@@ -705,7 +713,6 @@ static GCS_BACKEND_NAME_FN(gcomm_name)
     return name;
 }
 
-
 static GCS_BACKEND_OPEN_FN(gcomm_open)
 {
     GCommConn::Ref ref(backend);
@@ -719,7 +726,6 @@ static GCS_BACKEND_OPEN_FN(gcomm_open)
 
     try
     {
-        gcomm::Critical<Protonet> crit(conn.get_pnet());
         conn.connect(channel, bootstrap);
     }
     catch (Exception& e)
