@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2019 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2020 Codership Oy <info@codership.com>
  *
  * $Id$
  */
@@ -12,6 +12,10 @@
 #include <gu_logger.hpp>
 #include <gu_macros.hpp>
 #include <gu_unordered.hpp>
+#include <gu_lock.hpp>
+#include <gu_uuid.hpp>
+
+#include <wsrep_membership_service.h>
 
 #include <errno.h>
 
@@ -22,7 +26,9 @@ uint8_t     const GCS_VOTE_POLICY_DEFAULT(0);
 
 void gcs_group_register(gu::Config* cnf)
 {
-    cnf->add(GCS_VOTE_POLICY_KEY);
+    cnf->add(GCS_VOTE_POLICY_KEY,
+             gu::Config::Flag::read_only |
+             gu::Config::Flag::type_integer);
 }
 
 const char* gcs_group_state_str[GCS_GROUP_STATE_MAX] =
@@ -151,13 +157,16 @@ group_nodes_init (const gcs_group_t* group, const gcs_comp_msg_t* comp)
 }
 
 /* Free nodes array */
-static void
+#ifndef GCS_CORE_TESTING
+static
+#endif // GCS_CORE_TESTING
+void
 group_nodes_free (gcs_group_t* group)
 {
-    int i;
+    assert(group->memb_mtx_.locked());
 
     /* cleanup after disappeared members */
-    for (i = 0; i < group->num; i++) {
+    for (int i = 0; i < group->num; i++) {
         gcs_node_free (&group->nodes[i]);
     }
 
@@ -166,6 +175,7 @@ group_nodes_free (gcs_group_t* group)
     group->nodes  = NULL;
     group->num    = 0;
     group->my_idx = -1;
+    group->memb_epoch_ = GCS_SEQNO_ILL;
 }
 
 void
@@ -173,17 +183,18 @@ gcs_group_free (gcs_group_t* group)
 {
     if (group->my_name)    free ((char*)group->my_name);
     if (group->my_address) free ((char*)group->my_address);
-    group_nodes_free (group);
     delete group->vote_history;
+
+    gu::Lock lock(group->memb_mtx_);
+    group_nodes_free (group);
 }
 
 /* Reset nodes array without breaking the statistics */
 static inline void
 group_nodes_reset (gcs_group_t* group)
 {
-    int i;
-    /* reset recv_acts at the nodes */
-    for (i = 0; i < group->num; i++) {
+    /* reset recv_acts at the nodes (no need to protect with memb_mtx_) */
+    for (int i = 0; i < group->num; i++) {
         if (i != group->my_idx) {
             gcs_node_reset (&group->nodes[i]);
         }
@@ -225,8 +236,8 @@ group_redo_last_applied (gcs_group_t* group)
         assert( 0  < group->last_applied_proto_ver ||
                -1 == group->last_applied_proto_ver /* for unit tests */);
 
-        log_debug << "last_last_applied[" << n << "]: "
-                  << node->id << ", " << node->last_applied << ", "
+        log_debug << "last_last_applied[" << group->nodes[n].name << "]: "
+                  << node->id << ", " << node->last_applied << ", count: "
                   << (group_count_last_applied(*group, *node) ? "yes" : "no");
 
         /* NOTE: It is crucial for consistency that last_applied algorithm
@@ -281,12 +292,17 @@ group_redo_last_applied (gcs_group_t* group)
         group->last_node    = last_node;
     }
 
-    log_debug << "final last_applied: " << group->last_applied;
+    log_debug << "final last_applied on " << group->nodes[group->my_idx].name
+              << ": " << group->last_applied;
 }
 
 static void
 group_go_non_primary (gcs_group_t* group)
 {
+    assert(group->memb_mtx_.locked());
+
+    group->memb_epoch_ = group->act_id_;
+
     if (group->my_idx >= 0) {
         assert(group->num > 0);
         assert(group->nodes);
@@ -365,6 +381,8 @@ group_check_donor (gcs_group_t* group)
 static void
 group_post_state_exchange (gcs_group_t* group)
 {
+    assert(group->memb_mtx_.locked());
+
     const gcs_state_msg_t* states[group->num];
     gcs_state_quorum_t* quorum = &group->quorum;
     bool new_exchange = gu_uuid_compare (&group->state_uuid, &GU_UUID_NIL);
@@ -417,11 +435,11 @@ group_post_state_exchange (gcs_group_t* group)
             {
                 gu_fatal("Reversing history: %lld -> %lld, this member has "
                          "applied %lld more events than the primary component."
-                         "Data loss is possible. Aborting.",
+                         "Data loss is possible. Must abort.",
                          (long long)group->act_id_, (long long)quorum->act_id,
                          (long long)(group->act_id_ - quorum->act_id));
-                assert(0);
-                gu_abort();
+                group->state  = GCS_GROUP_INCONSISTENT;
+                return;
             }
             group->state      = GCS_GROUP_PRIMARY;
             group->act_id_    = quorum->act_id;
@@ -430,8 +448,10 @@ group_post_state_exchange (gcs_group_t* group)
             group->prim_uuid  = group->state_uuid;
             group->state_uuid = GU_UUID_NIL;
 
-            if (quorum->gcs_proto_ver >= 2) // see below for older version
+            if (quorum->gcs_proto_ver == 2) // see below for other versions
             {
+                /* version 2 was a mistake, but we can't eliminate this code
+                 * path for the sake of backward compatibility */
                 assert(quorum->last_applied >= 0);
                 group->last_applied = quorum->last_applied;
             }
@@ -459,11 +479,10 @@ group_post_state_exchange (gcs_group_t* group)
         GROUP_UPDATE_PROTO_VER(appl);
 #undef GROUP_UPDATE_PROTO_VER
 
-        if (quorum->gcs_proto_ver < 2) // see above for newer version
+        if (quorum->gcs_proto_ver != 2) // see above for version 2
         {
             group_redo_last_applied(group);
         }
-        // votes will be recounted on CC action creation
     }
     else {
         // non-primary configuration
@@ -609,6 +628,7 @@ gcs_group_handle_comp_msg (gcs_group_t* group, const gcs_comp_msg_t* comp)
         }
     }
     else {
+        gu::Lock lock(group->memb_mtx_);
         group_go_non_primary (group);
     }
 
@@ -629,17 +649,23 @@ gcs_group_handle_comp_msg (gcs_group_t* group, const gcs_comp_msg_t* comp)
         new_memb |= (old_idx == group->num);
     }
 
-    /* free old nodes array */
-    group_nodes_free (group);
+    {
+        gu::Lock lock(group->memb_mtx_);
 
-    group->my_idx = new_my_idx;
-    group->num    = new_nodes_num;
-    group->nodes  = new_nodes;
+        /* free old nodes array */
+        group_nodes_free (group);
 
-    assert(group->num > 0 || group->my_idx < 0);
-    assert(group->my_idx >= 0 || group->num == 0);
+        group->my_idx = new_my_idx;
+        group->num    = new_nodes_num;
+        group->nodes  = new_nodes;
+        group->memb_epoch_ = group->act_id_;
 
-    if (group->my_idx >= 0) group->nodes[group->my_idx].bootstrap = my_bootstrap;
+        assert(group->num > 0 || group->my_idx < 0);
+        assert(group->my_idx >= 0 || group->num == 0);
+
+        if (group->my_idx >= 0)
+            group->nodes[group->my_idx].bootstrap = my_bootstrap;
+    }
 
     if (gcs_comp_msg_primary(comp) || bootstrap) {
         /* TODO: for now pretend that we always have new nodes and perform
@@ -694,14 +720,6 @@ gcs_group_handle_uuid_msg  (gcs_group_t* group, const gcs_recv_msg_t* msg)
     return group->state;
 }
 
-static void group_print_state_debug(gcs_state_msg_t* state)
-{
-    size_t str_len = 1024;
-    char state_str[str_len];
-    gcs_state_msg_snprintf (state_str, str_len, state);
-    gu_info ("%s", state_str);
-}
-
 gcs_group_state_t
 gcs_group_handle_state_msg (gcs_group_t* group, const gcs_recv_msg_t* msg)
 {
@@ -722,10 +740,12 @@ gcs_group_handle_state_msg (gcs_group_t* group, const gcs_recv_msg_t* msg)
                          msg->sender_idx, gcs_state_msg_name(state));
                 gu_debug("%s", state_str);
 
-                if (gu_log_debug) group_print_state_debug(state);
-
-                gcs_node_record_state (&group->nodes[msg->sender_idx], state);
-                group_post_state_exchange (group);
+                {
+                    gu::Lock lock(group->memb_mtx_);
+                    group->memb_epoch_ = group->act_id_;
+                    gcs_node_record_state(&group->nodes[msg->sender_idx], state);
+                    group_post_state_exchange (group);
+                }
             }
             else {
                 gu_debug ("STATE EXCHANGE: stray state msg: " GU_UUID_FORMAT
@@ -736,8 +756,6 @@ gcs_group_handle_state_msg (gcs_group_t* group, const gcs_recv_msg_t* msg)
                           GU_UUID_ARGS(&group->state_uuid));
                 gu_debug ("%s", state_str);
 
-                if (gu_log_debug) group_print_state_debug(state);
-
                 gcs_state_msg_destroy (state);
             }
         }
@@ -815,9 +833,18 @@ gcs_group_handle_last_msg (gcs_group_t* group, const gcs_recv_msg_t* msg)
     // between threads.
     // assert (seqno >= group->last_applied);
 
-    gcs_node_set_last_applied (&group->nodes[msg->sender_idx], gtid.seqno());
+    {
+        gu::Lock lock(group->memb_mtx_);
+        group->memb_epoch_ = group->act_id_;
+        gcs_node_set_last_applied (&group->nodes[msg->sender_idx], gtid.seqno());
+    }
     assert(group->nodes[msg->sender_idx].last_applied >= 0);
 
+    log_debug << "Got last applied " << gtid.seqno() << " from "
+              << msg->sender_idx << " (" << group->nodes[msg->sender_idx].name
+              << "). Last node: " << group->last_node << " ("
+              << group->nodes[group->last_node].name << ")";
+
     if (msg->sender_idx == group->last_node   &&
         gtid.seqno()    >  group->last_applied) {
         /* node that was responsible for the last value, has changed it.
@@ -827,8 +854,8 @@ gcs_group_handle_last_msg (gcs_group_t* group, const gcs_recv_msg_t* msg)
         group_redo_last_applied (group);
 
         if (old_val < group->last_applied) {
-            gu_debug ("New COMMIT CUT %lld after %lld from %d",
-                      (long long)group->last_applied,
+            gu_debug ("New COMMIT CUT %lld on %d after %lld from %d",
+                      (long long)group->last_applied, group->my_idx,
                       (long long)gtid.seqno(), msg->sender_idx);
             return group->last_applied;
         }
@@ -1005,8 +1032,15 @@ gcs_group_handle_vote_msg (gcs_group_t* group, const gcs_recv_msg_t* msg)
         return ret;
     }
 
-    if (gtid.uuid() == group->group_uuid &&
-        gtid.seqno() > group->vote_result.seqno)
+    /* If either group-wide vote seqno or last applied are greater than the
+       request seqno, the vote has either happened already or there was no
+       need (i.e. all other members had a success). */
+    gcs_seqno_t const min_seqno =
+        group->quorum.gcs_proto_ver >= 4
+        ? std::max(group->last_applied, group->vote_result.seqno)
+        : group->vote_result.seqno;
+
+    if (gtid.uuid() == group->group_uuid && gtid.seqno() > min_seqno)
     {
         const char* const data
             (gcs::core::CodeMsg::serial_size() < msg->size ?
@@ -1019,7 +1053,8 @@ gcs_group_handle_vote_msg (gcs_group_t* group, const gcs_recv_msg_t* msg)
                  << gtid << ',' << gu::PrintBase<>(code) << ": "
                  << (code ? (data ? data : "(null)") : "Success");
 
-        gcs_node_set_vote (&sender, gtid.seqno(), code);
+        gcs_node_set_vote (&sender, gtid.seqno(), code,
+                           group->quorum.gcs_proto_ver);
 
         if (group_recount_votes(*group))
         {
@@ -1101,7 +1136,11 @@ gcs_group_handle_join_msg  (gcs_group_t* group, const gcs_recv_msg_t* msg)
     if (gu_unlikely(group_unserialize_code_msg(group, msg, gtid,code))) return 0;
 
     if (GCS_NODE_STATE_DONOR  == sender->status ||
-        GCS_NODE_STATE_JOINER == sender->status) {
+        GCS_NODE_STATE_JOINER == sender->status)
+    {
+        gu::Lock lock(group->memb_mtx_);
+        group->memb_epoch_ = group->act_id_;
+
         int j;
         gcs_node_t* peer      = NULL;
         const char* peer_id   = NULL;
@@ -1241,7 +1280,10 @@ gcs_group_handle_sync_msg  (gcs_group_t* group, const gcs_recv_msg_t* msg)
     if (GCS_NODE_STATE_JOINED == sender->status ||
         /* #454 - at this layer we jump directly from DONOR to SYNCED */
         (0 == group->last_applied_proto_ver &&
-         GCS_NODE_STATE_DONOR == sender->status)) {
+         GCS_NODE_STATE_DONOR == sender->status))
+    {
+        gu::Lock lock(group->memb_mtx_);
+        group->memb_epoch_ = group->act_id_;
 
         sender->status = GCS_NODE_STATE_SYNCED;
         sender->count_last_applied = group_count_arbitrator(*group, *sender);
@@ -1415,7 +1457,7 @@ group_for_each_donor_in_string (const gcs_group_t* const group,
          * that at least one of the nodes in the list will become available. */
         if (-EAGAIN != err) err = idx;
 
-        begin = end + 1; /* skip comma */
+        if (end) begin = end + 1; /* skip comma */
 
     } while (end != NULL);
 
@@ -1504,7 +1546,7 @@ group_find_ist_donor_by_name_in_string (
                 ret = idx;
             }
         }
-        begin = end + 1;
+        if (end) begin = end + 1;
     } while (end != NULL);
 
     if (ret == -1) {
@@ -1700,7 +1742,11 @@ group_select_donor (gcs_group_t* group,
                                          donor_string, donor_len, ist_gtid);
     }
 
-    if (donor_idx >= 0) {
+    if (donor_idx >= 0)
+    {
+        gu::Lock lock(group->memb_mtx_);
+        group->memb_epoch_ = group->act_id_;
+
         assert(donor_idx != joiner_idx || desync);
 
         gcs_node_t* const joiner = &group->nodes[joiner_idx];
@@ -1749,8 +1795,11 @@ group_select_donor (gcs_group_t* group,
 void
 gcs_group_ignore_action (gcs_group_t* group, struct gcs_act_rcvd* act)
 {
-//    if (act->act.type <= GCS_ACT_STATE_REQ) {
-    if (act->act.type <= GCS_ACT_CCHANGE) {
+    gu_debug("Ignoring action: buf: %p, len: %zd, type: %d, sender: %d, "
+             "seqno: %lld", act->act.buf, act->act.buf_len, act->act.type,
+             act->sender_idx, act->id);
+
+    if (gcs_act_in_cache(act->act.type)) {
         gcs_gcache_free (group->cache, act->act.buf);
     }
 
@@ -1774,8 +1823,11 @@ int
 gcs_group_handle_state_request (gcs_group_t*         group,
                                 struct gcs_act_rcvd* act)
 {
-    // pass only to sender and to one potential donor
-    const char* const donor_name    = (const char*)act->act.buf;
+    // 1. pass only to sender and to one potential donor
+    // 2. the message shall be stripped of donor preamble, hence const_cast
+    char* const donor_name =
+        (char*)gcs_gcache_get_rw_plaintext(group->cache,
+                                           const_cast<void*>(act->act.buf));
     size_t const     donor_name_len = strlen(donor_name) + 1;
     int              donor_idx      = -1;
     int const        joiner_idx     = act->sender_idx;
@@ -1786,31 +1838,57 @@ gcs_group_handle_state_request (gcs_group_t*         group,
     gu::GTID ist_gtid;
     int str_version = 1; // actually it's 0 or 1.
 
-    if (act->act.buf_len > (ssize_t)donor_name_len &&
-        donor_name[donor_name_len + 0] == 'V') {
+    if (act->act.buf_len > (ssize_t)(donor_name_len + 1) &&
+        donor_name[donor_name_len] == 'V') {
         str_version = (int)donor_name[donor_name_len + 1];
     }
 
     if (str_version >= 2) {
-        size_t offset(donor_name_len + 2);
+        ssize_t const ist_offset(donor_name_len + 2);
+        ssize_t const sst_offset(ist_offset + gu::GTID::serial_size());
 
         try
         {
-            offset = ist_gtid.unserialize(act->act.buf, act->act.buf_len,offset);
+            if (act->act.buf_len < sst_offset)
+            {
+                gu_throw_error(EINVAL) << "Request message too short: "
+                                       << act->act.buf_len << " < "
+                                       << sst_offset;
+            }
+
+            ssize_t const offset
+                (ist_gtid.unserialize(donor_name, act->act.buf_len,
+                                      ist_offset));
+            if (offset != sst_offset)
+            {
+                gu_throw_error(EINVAL) << "Actual SST offset " << offset
+                                       << " does not match expected "
+                                       << sst_offset;
+            }
         }
-        catch (gu::Exception& e) {
-            log_warn << "Malformed state transfer request: " << e.what()
-                     << " Ignoring";
-            gcs_group_ignore_action(group, act);
-            return 0;
+        catch (gu::Exception& e)
+        {
+            if (group->my_idx == joiner_idx)
+            {
+                log_fatal << "Failed to form State Transfer Request: "
+                          << e.what();
+                act->id = -ENOTRECOVERABLE;
+                return act->act.buf_len;
+            }
+            else
+            {
+               log_warn << "Malformed State Transfer Request: " << e.what()
+                        << " Ignoring";
+                gcs_group_ignore_action(group, act);
+                return 0;
+            }
         }
 
         // change act.buf's content to original version.
         // and it's safe to change act.buf_len
-        ::memmove((char*)act->act.buf + donor_name_len,
-                  (char*)act->act.buf + offset,
-                  act->act.buf_len - offset);
-        act->act.buf_len -= offset - donor_name_len;
+        ::memmove(donor_name + donor_name_len, donor_name + sst_offset,
+                  act->act.buf_len - sst_offset);
+        act->act.buf_len -= sst_offset - donor_name_len;
     }
 
     assert (GCS_ACT_STATE_REQ == act->act.type);
@@ -1820,16 +1898,27 @@ gcs_group_handle_state_request (gcs_group_t*         group,
         const char* joiner_status_string = gcs_node_state_to_str(joiner_status);
 
         if (group->my_idx == joiner_idx) {
-            gu_error ("Requesting state transfer while in %s. "
-                      "Ignoring.", joiner_status_string);
-            act->id = -ECANCELED;
+            if (joiner_status >= GCS_NODE_STATE_JOINED)
+            {
+                gu_warn ("Requesting state transfer while in %s. "
+                         "Ignoring.", joiner_status_string);
+                act->id = -ECANCELED;
+            }
+            else
+            {
+                /* The node can't send two STRs in a row */
+                assert(joiner_status == GCS_NODE_STATE_JOINER);
+                gu_fatal("Requesting state transfer while in %s. "
+                         "Internal program error.", joiner_status_string);
+                act->id = -ENOTRECOVERABLE;
+            }
             return act->act.buf_len;
         }
         else {
-            gu_error ("Member %d.%d (%s) requested state transfer, "
-                      "but its state is %s. Ignoring.",
-                      joiner_idx, group->nodes[joiner_idx].segment, joiner_name,
-                      joiner_status_string);
+            gu_warn ("Member %d.%d (%s) requested state transfer, "
+                     "but its state is %s. Ignoring.",
+                     joiner_idx, group->nodes[joiner_idx].segment, joiner_name,
+                     joiner_status_string);
             gcs_group_ignore_action (group, act);
             return 0;
         }
@@ -1848,9 +1937,7 @@ gcs_group_handle_state_request (gcs_group_t*         group,
     }
     else if (group->my_idx == donor_idx) {
         act->act.buf_len -= donor_name_len;
-        memmove (*(void**)&act->act.buf,
-                 ((char*)act->act.buf) + donor_name_len,
-                 act->act.buf_len);
+        memmove (donor_name, donor_name + donor_name_len, act->act.buf_len);
         // now action starts with request, like it was supplied by application,
         // see gcs_request_state_transfer()
     }
@@ -1861,6 +1948,7 @@ gcs_group_handle_state_request (gcs_group_t*         group,
     // This may be ugly, well, any ideas?
     act->id = donor_idx;
 
+    gcs_gcache_drop_plaintext(group->cache, act->act.buf);
     return act->act.buf_len;
 }
 
@@ -1896,12 +1984,12 @@ gcs_group_act_conf (gcs_group_t*         group,
                 conf.vote_res   = group->vote_result.res;
             }
         }
-        conf.seqno      = group->act_id_;
-    } else {
+    }
+    else {
         assert(GCS_GROUP_NON_PRIMARY == group->state);
-        conf.seqno      = GCS_SEQNO_ILL;
     }
 
+    conf.seqno          = group->act_id_;
     conf.conf_id        = group->conf_id;
     conf.repl_proto_ver = group->quorum.repl_proto_ver;
     conf.appl_proto_ver = group->quorum.appl_proto_ver;
@@ -1935,10 +2023,13 @@ gcs_group_act_conf (gcs_group_t*         group,
     rcvd->act.buf_len = conf.write(&tmp); // throws when fails
 #ifndef GCS_FOR_GARB
     /* copy CC event to gcache for IST */
-    rcvd->act.buf = gcache_malloc(group->cache, rcvd->act.buf_len);
+    void* ptx;
+    rcvd->act.buf = gcache_malloc(group->cache, rcvd->act.buf_len, &ptx);
     if (rcvd->act.buf)
     {
-        memcpy(const_cast<void*>(rcvd->act.buf), tmp, rcvd->act.buf_len);
+        assert(ptx);
+        memcpy(ptx, tmp, rcvd->act.buf_len);
+        gcache_drop_plaintext(group->cache, rcvd->act.buf);
         rcvd->id = group->my_idx; // passing own index in seqno_g
     }
     else
@@ -2026,8 +2117,7 @@ gcs_group_param_set(gcs_group_t& group,
     return 1;
 }
 
-void
-gcs_group_get_status (const gcs_group_t* group, gu::Status& status)
+int gcs_group_get_status (const gcs_group_t* group, gu::Status& status)
 {
     int desync_count; // make sure it is not initialized
 
@@ -2043,5 +2133,199 @@ gcs_group_get_status (const gcs_group_t* group, gu::Status& status)
     }
 
     status.insert("desync_count", gu::to_string(desync_count));
+
+    return 0;
+}
+
+void
+gcs_group_get_membership(const gcs_group_t&        group,
+                         wsrep_allocator_cb const  alloc,
+                         struct wsrep_membership** memb)
+{
+    if (!alloc)
+    {
+        gu_throw_error(EINVAL) << "No allocator for membership return value";
+    }
+
+    gu::Lock lock(group.memb_mtx_);
+
+    size_t const memb_size
+        (sizeof(struct wsrep_membership) +
+         (group.num - 1)*sizeof(struct wsrep_member_info_ext));
+
+    *memb = static_cast<struct wsrep_membership*>(alloc(memb_size));
+
+    if (!*memb)
+    {
+        gu_throw_error(ENOMEM) << "Could not allocate " << memb_size
+                               << " bytes for membership struct.";
+    }
+
+    ::memset(*memb, 0, memb_size);
+
+    struct wsrep_membership& m(**memb);
+
+    m.group_uuid = group.group_uuid;
+    m.updated = group.memb_epoch_;
+
+    switch (group.state)
+    {
+    case GCS_GROUP_NON_PRIMARY:
+        m.state = WSREP_VIEW_NON_PRIMARY;
+        break;
+    case GCS_GROUP_WAIT_STATE_UUID:
+    case GCS_GROUP_WAIT_STATE_MSG:
+    case GCS_GROUP_PRIMARY:
+    case GCS_GROUP_INCONSISTENT:
+        m.state = WSREP_VIEW_PRIMARY;
+        break;
+    case GCS_GROUP_STATE_MAX:
+        m.state = WSREP_VIEW_DISCONNECTED;
+    }
+
+    m.num = group.num;
+
+    for (size_t i(0); i < m.num; ++i)
+    {
+        struct wsrep_member_info_ext& mn(m.members[i]);
+        const struct gcs_node& gn(group.nodes[i]);
+
+        gu_uuid_t uuid;
+        gu_uuid_scan(gn.id, sizeof(gn.id), &uuid);
+        mn.base.id = uuid;
+
+        ::snprintf(mn.base.name, sizeof(mn.base.name) - 1,//leave last byte for 0
+                   "%s", group.nodes[i].name);
+        ::snprintf(mn.base.incoming, sizeof(mn.base.incoming) - 1,
+                   "%s", group.nodes[i].inc_addr);
+
+        mn.last_committed = group.nodes[i].last_applied;
+
+        switch(group.nodes[i].status)
+        {
+        case GCS_NODE_STATE_NON_PRIM:
+        case GCS_NODE_STATE_PRIM:
+            mn.status = WSREP_MEMBER_UNDEFINED;
+            break;
+        case GCS_NODE_STATE_JOINER:
+            mn.status = WSREP_MEMBER_JOINER;
+            break;
+        case GCS_NODE_STATE_DONOR:
+            mn.status = WSREP_MEMBER_DONOR;
+            break;
+        case GCS_NODE_STATE_JOINED:
+            mn.status = WSREP_MEMBER_JOINED;
+            break;
+        case GCS_NODE_STATE_SYNCED:
+            mn.status = WSREP_MEMBER_SYNCED;
+            break;
+        case GCS_NODE_STATE_MAX:
+            mn.status = WSREP_MEMBER_ERROR;
+        }
+    }
+}
+int gcs_group_fetch_pfs_info(const gcs_group_t*  group,
+                             wsrep_node_info_t** nodes_arg,
+                             uint32_t*           size,
+                             int32_t*            my_index,
+                             uint32_t            max_version)
+{
+    int num = (int) group->num;
+    if (num <= 0)
+    {
+        return -ENOTCONN;
+    }
+ 
+    wsrep_node_info_t* entries = static_cast<wsrep_node_info_t*>(
+        gu_malloc(sizeof(wsrep_node_info_t) * num)
+    );
+    if (!entries)
+    {
+        gu_warn("Failed to allocate nodes information structure");
+        return -ENOMEM;
+    }
+
+    *nodes_arg = entries;
+    *size = num;
+    *my_index = (int) group->my_idx;
+
+    int i;
+    for (i = 0; i < num; i++)
+    {
+        const gcs_node_t& node(group->nodes[i]);
+
+        entries[i].wsrep_version = max_version;
+        entries[i].wsrep_local_index = i;
+
+        memcpy(entries[i].wsrep_node_id, node.id, WSREP_UUID_STR_LEN);
+        entries[i].wsrep_node_id[WSREP_UUID_STR_LEN] = 0;
+
+        strncpy(entries[i].wsrep_host_name, node.name,
+                WSREP_HOSTNAME_LENGTH);
+        entries[i].wsrep_host_name[WSREP_HOSTNAME_LENGTH] = 0;
+
+        gu_uuid_print(&group->group_uuid,
+                      entries[i].wsrep_cluster_state_uuid,
+                      WSREP_UUID_STR_LEN + 1);
+
+        gu_uuid_print(&group->state_uuid,
+                      entries[i].wsrep_local_state_uuid,
+                      WSREP_UUID_STR_LEN + 1);
+
+        strncpy(entries[i].wsrep_status,
+                gcs_node_state_to_str(node.status),
+                WSREP_STATUS_LENGTH);
+        entries[i].wsrep_status[WSREP_STATUS_LENGTH] = 0;
+
+        entries[i].wsrep_segment = node.segment;
+
+        entries[i].wsrep_last_applied = node.last_applied;
+        entries[i].wsrep_last_committed = 0;
+        entries[i].wsrep_replicated = 0;
+        entries[i].wsrep_replicated_bytes = 0;
+        entries[i].wsrep_received = 0;
+        entries[i].wsrep_received_bytes = 0;
+        entries[i].wsrep_local_bf_aborts = 0;
+        entries[i].wsrep_local_commits = 0;
+        entries[i].wsrep_local_cert_failures = 0;
+        entries[i].wsrep_apply_window = 0;
+        entries[i].wsrep_commit_window = 0;
+    }
+
+    return 0;
+}
+
+int gcs_group_fetch_pfs_stat(const gcs_group_t*  group,
+                             wsrep_node_stat_t** nodes_arg,
+                             uint32_t*           size,
+                             int32_t*            my_index,
+                             uint32_t            max_version)
+{
+    int num = (int) group->num;
+    int my_idx = (int) group->my_idx;
+    if (num > 0 && my_idx >= 0)
+    {
+        wsrep_node_stat_t* entry = static_cast<wsrep_node_stat_t*>(
+            gu_malloc(sizeof(wsrep_node_stat_t))
+        );
+        if (!entry)
+        {
+            gu_warn("Failed to allocate node statistics structure");
+            return -ENOMEM;
+        }
+        *nodes_arg = entry;
+        *size = 1;
+        *my_index = 0;
+        entry->wsrep_version = max_version;
+        entry->wsrep_local_index = my_idx;
+        const gcs_node_t& node(group->nodes[my_idx]);
+        strncpy(entry->wsrep_node_id, node.id, WSREP_UUID_STR_LEN);
+        entry->wsrep_node_id[WSREP_UUID_STR_LEN] = 0;
+        return 0;
+    }
+    else
+    {
+        return -ENOTCONN;
+    }
 }
 
