@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2025 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2020 Codership Oy <info@codership.com>
  *
  * $Id$
  */
@@ -7,7 +7,6 @@
 #include "gcs_node.hpp"
 #include "gcs_state_msg.hpp"
 #include <stdlib.h>
-#include <cinttypes>
 
 #include <gu_utils.hpp> // gu::PrintBase
 
@@ -21,8 +20,7 @@ gcs_node_init (gcs_node_t* const node,
                int const gcs_proto_ver,
                int const repl_proto_ver,
                int const appl_proto_ver,
-               gcs_segment_t const segment,
-               bool const stateless)
+               gcs_segment_t const segment)
 {
     assert(strlen(id) > 0);
     assert(strlen(id) < sizeof(node->id));
@@ -41,7 +39,6 @@ gcs_node_init (gcs_node_t* const node,
     node->repl_proto_ver = repl_proto_ver;
     node->appl_proto_ver = appl_proto_ver;
     node->segment        = segment;
-    node->stateless      = stateless;
 }
 
 /*! Move data from one node object to another */
@@ -195,10 +192,10 @@ gcs_node_update_status (gcs_node_t* node, const gcs_state_quorum_t* quorum)
             else {
                 // gap in sequence numbers, needs a snapshot, demote status
                 if (node->status > GCS_NODE_STATE_PRIM) {
-                    gu_info("'%s' demoted %s->PRIMARY due to gap in history: "
-                            "%" PRId64 " - %" PRId64,
-                            node->name, gcs_node_state_to_str(node->status),
-                            node_act_id, quorum->act_id);
+                    gu_info ("'%s' demoted %s->PRIMARY due to gap in history: "
+                             "%lld - %lld",
+                             node->name, gcs_node_state_to_str(node->status),
+                             node_act_id, quorum->act_id);
                 }
                 node->status = GCS_NODE_STATE_PRIM;
             }
@@ -278,8 +275,8 @@ gcs_node_update_status (gcs_node_t* node, const gcs_state_quorum_t* quorum)
      * subsequent configuration changes. */
     node->bootstrap = false;
 
-    node->stateless = (gcs_state_msg_flags (node->state_msg) &
-                       GCS_STATE_FSTATELESS);
+    node->arbitrator = (gcs_state_msg_flags (node->state_msg) &
+                        GCS_STATE_ARBITRATOR);
 }
 
 void
@@ -300,5 +297,5 @@ gcs_node_print(std::ostream& os, const gcs_node_t& node)
        << "status:\t " << gcs_node_state_to_str(node.status) << '\n'
        << "segment:  " << int(node.segment) << '\n'
        << "bootstrp: " << (node.bootstrap ? "YES" : "NO") << '\n'
-       << "arbitr: "   << (node.stateless ? "YES" : "NO");
+       << "arbitr: "   << (node.arbitrator ? "YES" : "NO");
 }
